
import torch
from torch import nn
import numpy as np


class WienerDeconv(nn.Module):
    def __init__(self, params):
        super().__init__()
        self.inverse_filt = (params['test']['model'] == 'inverse')

    @torch.no_grad()
    # adapted from Jared's HW3 implementation
    def forward(self, imgs, cFT, sigma):
        '''
        imgs, cFT, and sigma are all batched
        '''
        B = imgs.shape[0]
        unfilt = torch.fft.fft2(imgs)
        cFT = cFT.repeat(B, 1, 1).unsqueeze(1)

        if self.inverse_filt:
            factor = 1.0 / cFT
        else:
            invSNR = sigma / imgs.mean(dim=(-1,-2), keepdim=True)
            H, H2 = cFT, torch.abs(cFT) ** 2
            factor = H2 / (H * (H2 + invSNR))
        return torch.clip(torch.real(torch.fft.ifft2(unfilt * factor)), 0, 1)
